package com.cnbm.processInspection.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.baomidou.mybatisplus.core.toolkit.ObjectUtils;
import com.cnbm.basic.dto.ProductFeaturesDTO;
import com.cnbm.basic.entity.Machine;
import com.cnbm.basic.entity.Product;
import com.cnbm.basic.entity.Shift;
import com.cnbm.basic.entity.WorkingProcedure;
import com.cnbm.basic.service.impl.MachineServiceImpl;
import com.cnbm.basic.service.impl.ProductServiceImpl;
import com.cnbm.basic.service.impl.ShiftServiceImpl;
import com.cnbm.basic.service.impl.WorkingProcedureServiceImpl;
import com.cnbm.common.page.PageData;
import com.cnbm.common.service.impl.CrudServiceImpl;
import com.cnbm.common.spc.util.DataUtils;
import com.cnbm.common.utils.ConvertUtils;
import com.cnbm.influx.config.InfluxClient;
import com.cnbm.influx.param.PageInfo;
import com.cnbm.influx.param.QueryDataParam;
import com.cnbm.influx.param.Range;
import com.cnbm.influx.param.Tag;
import com.cnbm.influx.template.Event;
import com.cnbm.processInspection.dto.InspectionSampleDTO;
import com.cnbm.processInspection.dto.InspectionSheetDTO;
import com.cnbm.processInspection.entity.InspectionSheet;
import com.cnbm.processInspection.mapper.InspectionSheetMapper;
import com.cnbm.processInspection.service.IInspectionSheetService;
import com.google.gson.JsonObject;
import com.influxdb.query.FluxRecord;
import com.influxdb.query.FluxTable;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.bind.annotation.RequestBody;

import javax.annotation.Resource;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 检验单 表
 *
 * @author why
 * @since 2022-08-17
 */
@Service
public class InspectionSheetServiceImpl extends CrudServiceImpl<InspectionSheetMapper, InspectionSheet, InspectionSheetDTO> implements IInspectionSheetService {

    @Autowired
    private ProductServiceImpl productService;

    @Autowired
    private WorkingProcedureServiceImpl workingProcedureService;

    @Autowired
    private MachineServiceImpl machineService;

    @Autowired
    private ShiftServiceImpl shiftService;

    @Override
    public QueryWrapper<InspectionSheet> getWrapper(Map<String, Object> params) {
        Long id = (Long) params.get("id");
        String inspectionSite = (String) params.get("inspectionSite");
        String orderNumber = (String) params.get("orderNumber");
        String batchNumber = (String) params.get("batchNumber");

        QueryWrapper<InspectionSheet> wrapper = new QueryWrapper<>();
        wrapper.like(ObjectUtils.isNotNull(id), "id", id);
        wrapper.like(StringUtils.isNotBlank(inspectionSite), "inspection_stage", inspectionSite);
        wrapper.like(StringUtils.isNotBlank(orderNumber), "order_number", orderNumber);
        wrapper.like(StringUtils.isNotBlank(batchNumber), "batch_number", batchNumber);
        wrapper.eq(ObjectUtils.isNotNull(params.get("productId")), "product_id", params.get("productId"));
        wrapper.eq(ObjectUtils.isNotNull(params.get("workingProcedureId")), "working_procedure_id", params.get("workingProcedureId"));
        wrapper.eq(ObjectUtils.isNotNull(params.get("machineId")), "machine_id", params.get("machineId"));
        return wrapper;
    }

    @Override
    public PageData<InspectionSheetDTO> page(Map<String, Object> params) {
        IPage<InspectionSheet> page = baseDao.selectPage(
                getPage(params, "id", true),
                getWrapper(params)
        );
        PageData<InspectionSheetDTO> inspectionSheetDTOPageData = getPageData(page, InspectionSheetDTO.class);

        if (!inspectionSheetDTOPageData.getList().isEmpty()) {
            for (InspectionSheetDTO dto : inspectionSheetDTOPageData.getList()) {
                Product product = productService.selectById(dto.getProductId());
                WorkingProcedure workingProcedure = workingProcedureService.selectById(dto.getWorkingProcedureId());
                Machine machine = machineService.selectById(dto.getMachineId());
                Shift shift = shiftService.selectById(dto.getShiftId());

                String productName = null;
                String workingProcedureName = null;
                String machineName = null;
                String shiftName = null;

                if (product != null) {
                    productName = product.getName();
                }
                if (workingProcedure != null) {
                    workingProcedureName = workingProcedure.getName();
                }
                if (machine != null) {
                    machineName = machine.getName();
                }
                if (shift != null) {
                    shiftName = shift.getName();
                }

                dto.setProductName(productName);
                dto.setWorkingProcedureName(workingProcedureName);
                dto.setMachineName(machineName);
                dto.setShiftName(shiftName);
            }
        }
        return inspectionSheetDTOPageData;
    }

    @Override
    public InspectionSheetDTO get(Long id) {
        InspectionSheet entity = baseDao.selectById(id);
        InspectionSheetDTO dto = ConvertUtils.sourceToTarget(entity, InspectionSheetDTO.class);
        WorkingProcedure workingProcedure = workingProcedureService.selectById(dto.getWorkingProcedureId());
        if(workingProcedure!=null){
            dto.setWorkingProcedureName(workingProcedure.getName());
        }
        return dto;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public InspectionSheet saveSheet(InspectionSheetDTO dto) {
        // 验证是否有检验参数
        Map<String, Object> params = new HashMap<String, Object>();
        params.put("productId", dto.getProductId());
        params.put("workingProcedureId", dto.getWorkingProcedureId());
        params.put("inspectionStage", dto.getInspectionStage());
        List<ProductFeaturesDTO> inspectionSheetFeaturesList = getInspectionSheetFeaturesList(params);
        if (inspectionSheetFeaturesList != null && inspectionSheetFeaturesList.size() != 0) {
            //分组样本数=样本大小=检验特性分组数的最大值
            Integer numbersOfSamples = inspectionSheetFeaturesList.stream().max(Comparator.comparing(ProductFeaturesDTO::getSampleSize)).get().getSampleSize();
            if (numbersOfSamples != null) {
                dto.setNumberOfGroupedSamples(numbersOfSamples);
                dto.setNumberOfSamples(numbersOfSamples);
            }
        }
        /*
        if (inspectionSheetFeaturesList == null || inspectionSheetFeaturesList.size() == 0) {
            throw new RuntimeException("没有发现检验参数");
        } else {
            //分组样本数=样本大小=检验特性分组数的最大值
            Integer numbersOfSamples = inspectionSheetFeaturesList.stream().max(Comparator.comparing(ProductFeaturesDTO::getSampleSize)).get().getSampleSize();
            if (numbersOfSamples != null) {
                dto.setNumberOfGroupedSamples(numbersOfSamples);
                dto.setNumberOfSamples(numbersOfSamples);
            }
        }
        */
        //InspectionSheet entity = ConvertUtils.sourceToTarget(dto, InspectionSheet.class);
        InspectionSheet entity =  new InspectionSheet();
        BeanUtils.copyProperties(dto, entity);
        return entity;
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void update(InspectionSheetDTO dto) {
        InspectionSheet entity = ConvertUtils.sourceToTarget(dto, InspectionSheet.class);
        updateById(entity);
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void delete(Long[] ids) {
        //删除
        deleteBatchIds(Arrays.asList(ids));
    }

    @Resource
    private InspectionSheetMapper mapper;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public List<InspectionSheetDTO> list(Map<String, Object> params) {
        List<InspectionSheetDTO> list = mapper.list(params);
        return list;
    }

    @Override
    public List<InspectionSampleDTO> getFluxParamList(Long id) {
        List<Event> eventList = new ArrayList<>();
        InspectionSheetDTO dto = get(id);
        if (dto != null && dto.getWorkingProcedureName()!=null) {
            QueryDataParam queryDataParam = new QueryDataParam();
            String inspectionSheetId = id.toString();
            String workingProcedureName = dto.getWorkingProcedureName();
            queryDataParam.setMeasurement(workingProcedureName);
            //queryDataParam.setInspectionSheetId(inspectionSheetId);
            queryDataParam.setBucket("qgs-bucket");
            queryDataParam.setTag(new Tag("inspectionSheetId", inspectionSheetId));
            List<String> dropNames = new ArrayList<>();
            dropNames.add("transationId");
            queryDataParam.setDropedTagNames(dropNames);
            queryDataParam.setRange(new Range(DataUtils.getBeforeDate(999).toInstant(), DataUtils.getAfterDate(999).toInstant()));
            queryDataParam.setPageInfo(new PageInfo(1, 10));

            List<FluxTable> fluxTableList = InfluxClient.Client.query(queryDataParam);
            for (FluxTable fluxTable : fluxTableList) {
                List<FluxRecord> records = fluxTable.getRecords(); // 获取实际的记录操作
                for (FluxRecord fluxRecord : records) {
                    Instant time = fluxRecord.getTime();
                    String argName = (String) fluxRecord.getValueByKey("argName");
                    String argValue = (String)fluxRecord.getValueByKey("_value");
                    String sampleNo = (String) fluxRecord.getValueByKey("sampleNumber");
                    eventList.add(newEvent(time, inspectionSheetId, argName, argValue, sampleNo));
                }
            }
        }
        List<InspectionSampleDTO> list = new ArrayList<>();
        if (eventList != null || eventList.size() > 0) {
            //根据样本号分组
            Map<String, List<Event>> map = eventList.stream().collect(Collectors.groupingBy(Event::getSampleNumber));
            for (Map.Entry<String, List<Event>> entry : map.entrySet()) {
                String sampleNumber = entry.getKey();
                List<Event> events = entry.getValue();

                InspectionSampleDTO sampleDTO = new InspectionSampleDTO();
                sampleDTO.setSampleNumber(sampleNumber);
                //sampleDTO.setSampleTime(LocalDateTime.ofInstant(events.get(0).getTime(), ZoneId.systemDefault()));
                sampleDTO.setSampleTime(events.get(0).getTime());
                sampleDTO.setBatchNum(dto.getBatchNumber());
                sampleDTO.setInspectionSheetId(events.get(0).getInspectionSheetId());

                JSONObject jsonObject = new JSONObject();
                for(Event event:events){
                    jsonObject.put(event.getArgName(),event.getArgValue());
                }
                sampleDTO.setJsonData(jsonObject.toString());
                //System.out.println(entry);
                list.add(sampleDTO);
            }
        }
        return list;
    }

    @Override
    public void saveFluxParamList(InspectionSampleDTO[] lists) {
        for (InspectionSampleDTO dto : lists) {
            //String jsonData = {"workingProcedureName":"test","inspectionSheetId":"116","param1":"0.47","param2":"0.687","param2":"0.53"};
            String workingProcedureName = dto.getWorkingProcedureName();
            String inspectionSheetId = dto.getInspectionSheetId();
//            String batchNum = dto.getBatchNum();
            String sampleNumber = dto.getSampleNumber();
            String jsonData = dto.getJsonData();
            JSONObject json = JSON.parseObject(jsonData);
//            LocalDateTime sampleTime = dto.getSampleTime();
//            Instant eventTime = sampleTime.toInstant(ZoneOffset.UTC);
            Instant eventTime = dto.getSampleTime();
            List<Event> list = new ArrayList<>();
            for (Map.Entry entry : json.entrySet()) {
                String key = entry.getKey().toString();
                String v = entry.getValue().toString();
//                Double v = null;
//                try {
//                    v= Double.valueOf(entry.getValue().toString());
//                }catch (Exception e){
//                }
                list.add(newEvent(eventTime, inspectionSheetId, key, v, sampleNumber));
            }
            InfluxClient.Client.batchInsert(list, workingProcedureName);
        }
//        //样本数据更新后 计算检验单缺陷数不良数
//        System.out.println("--------------------------------------");
//        calculate(Long.valueOf(lists[0].getInspectionSheetId()));
    }

    private Event newEvent(Instant time, String inspectionSheetId, String argName, String argValue, String sampleNo) {
        Event event = new Event();
        event.setInspectionSheetId(inspectionSheetId);
        event.setTime(time);
//        event.setBatchNum(batchNum);
        event.setSampleNumber(sampleNo);
        event.setArgName(argName);
        event.setArgValue(argValue);
//        if (Objects.equals(argValue, "")){
//            event.setArgValue("");
//        }else if (argValue != null) {
//            event.setArgValue(argValue.toString());
//        }
        return event;
    }

//    @Override
//    public Integer getNumberOfSamples(Map<String, Object> params) {
//        return mapper.getNumberOfSamples(params);
//    }

    @Override
    public List<ProductFeaturesDTO> getInspectionSheetFeaturesList(Map<String, Object> params) {
        return mapper.getInspectionSheetFeaturesList(params);
    }

    //计算并更新检验单缺陷数不良数
    @Override
    public void calculate(Long inspectionSheetId) {
        List<InspectionSampleDTO> sampleDTOList = getFluxParamList(inspectionSheetId);
        if (sampleDTOList != null || sampleDTOList.size() > 0) {
            //缺陷
            Integer numberOfDefects = 0;
            //不良
            Integer defectiveQuantity = 0;

            //获取的产品特性参照值
            InspectionSheetDTO dto = get(inspectionSheetId);
            Map<String, Object> params2 = new HashMap<>();
            params2.put("productId", dto.getProductId());
            params2.put("workingProcedureId", dto.getWorkingProcedureId());
            params2.put("inspectionStage", dto.getInspectionStage());
            List<ProductFeaturesDTO> featuresList = getInspectionSheetFeaturesList(params2);
            //循环每个样本组
            for (InspectionSampleDTO sampleDTO : sampleDTOList) {
                //该样本的缺陷数
                Integer eventDefects = 0;
                //该样本的属性列表
                String jsonData = sampleDTO.getJsonData();
                JSONObject json = JSON.parseObject(jsonData);
                for (ProductFeaturesDTO feature : featuresList) {
                    String v = (String)json.get(feature.getName());
                    if (!Objects.equals(v, "") && v != null) {
                        //1 计量型；2 计数型
                        if (feature.getType() == 1) {
                            Double featureValue = Double.valueOf(v);
                            if (featureValue > feature.getUsl() || featureValue < feature.getLsl()) {
                                eventDefects = eventDefects + 1;
                            }
                        }
                        if (feature.getType() == 2) {
                            Integer featureValue = Integer.valueOf(v);
                            if (featureValue == 1) {
                                eventDefects = eventDefects + 1;
                            }
                        }
                    }
                    if (eventDefects != 0) {
                        numberOfDefects = numberOfDefects + eventDefects;
                        defectiveQuantity = defectiveQuantity + 1;
                    }
                }
            }
            InspectionSheetDTO updateDto = new InspectionSheetDTO();
            updateDto.setId(inspectionSheetId);
            updateDto.setNumberOfDefects(numberOfDefects);
            updateDto.setDefectiveQuantity(defectiveQuantity);
            update(updateDto);
        }

    }
        /*
        //获取该检验单的所有样本数据
        List<Event> eventList = getFluxParamList(inspectionSheetId);
        if (eventList != null || eventList.size() > 0) {
            //根据样本号分组
            Map<String, List<Event>> map = eventList.stream().collect(Collectors.groupingBy(Event::getSampleNumber));
            //获取的产品特性参照值
            InspectionSheetDTO dto = get(inspectionSheetId);
            Map<String, Object> params2 = new HashMap<>();
            params2.put("productId", dto.getProductId());
            params2.put("workingProcedureId", dto.getWorkingProcedureId());
            params2.put("inspectionStage", dto.getInspectionStage());
            //缺陷
            Integer numberOfDefects = 0;
            //不良
            Integer defectiveQuantity = 0;

            List<ProductFeaturesDTO> featuresList = getInspectionSheetFeaturesList(params2);
            //循环每个样本组
            for (Map.Entry<String, List<Event>> entry : map.entrySet()) {
                //该样本的缺陷数
                Integer eventDefects = 0;
                //该样本的属性列表
                List<Event> eventList1 = entry.getValue();
                for (Event event : eventList1) {
                    String featureName = event.getArgName();
                    ProductFeaturesDTO feature = featuresList.stream().filter(u -> u.getName().equals(featureName)).findAny().orElse(null);
                    if (feature != null && !Objects.equals(event.getArgValue(), "") && event.getArgValue() != null) {
                        //1 计量型；2 计数型
                        if (feature.getType() == 1) {
                            Double featureValue = Double.valueOf(event.getArgValue());
                            if (featureValue > feature.getUsl() || featureValue < feature.getLsl()) {
                                eventDefects = eventDefects + 1;
                            }
                        }
                        if (feature.getType() == 2) {
                            Integer featureValue = Integer.valueOf(event.getArgValue());
                            if (featureValue == 1) {
                                eventDefects = eventDefects + 1;
                            }
                        }
                    }
                    if (eventDefects != 0) {
                        numberOfDefects = numberOfDefects + eventDefects;
                        defectiveQuantity = defectiveQuantity + 1;
                    }
                }
            }
            InspectionSheetDTO updateDto = new InspectionSheetDTO();
            updateDto.setId(inspectionSheetId);
            updateDto.setNumberOfDefects(numberOfDefects);
            updateDto.setDefectiveQuantity(defectiveQuantity);
            update(updateDto);
        }
    }*/
}